const isStringOkey = (string, isJSX) => {
  const regex = /((^[A-Z]{1}.*?)|(\w*?\s\w*?.*?))/;
  const regexForMinMax = /(min-width)|(max-width)/;
  const regexForTwoUpperCaseInARow = /^[A-Z][a-z]*?[A-Z]/;
  const regexForPlural = /plural/;
  const regexForRem = /([0-9]rem)|([0-9]em)|([0-9]px)/;
  const regexForWrapper = /{*}/;
  const regexCSSRule = /:.{1,};/;
  const regexForLetter = /[a-z]/;
  const regexForObject = /\[object/;
  const regexForHTML = /<[a-z]/;
  const regexDate = /DD|MM|YY|HH|mm:ss/;
  const blackListRegex = [
    regexForMinMax,
    regexForPlural,
    regexForWrapper,
    regexForRem,
    regexCSSRule,
    regexForObject,
    regexForHTML,
    /Bearer/,
    regexDate
  ];

  return (
    regex.test(string) &&
    regexForLetter.test(string) &&
    (isJSX ||
      !(regexForTwoUpperCaseInARow.test(string) && !/\s/.test(string))) &&
    (isJSX ||
      !blackListRegex.some(regex => {
        return regex.test(string);
      }))
  );
};

const blackListOfAttributesToBeTransformed = [
  "defaultMessage",
  "p",
  "m",
  "width",
  "height",
  "flex",
  "minHeight",
  "maxHeight",
  "border",
  "borderTop",
  "borderRight",
  "borderBottom",
  "borderLeft",
  "lg",
  "md",
  "all",
  "xs",
  "xl",
  "xxl",
  "breakpoint",
  "sm",
  "viewBox",
  "d",
  "rel",
  "srcSet",
  "className",
  "Trans",
  "backgroundPosition",
  "backgroundSize",
  "id",
  "data-name",
  "preserveAspectRatio"
];

const whiteListOfCallExpressions = [
  "show",
  "validation",
  "getErrorMessage",
  "handleClientSideError",
  "renderTooltip"
];

const blackListOfFileNames = [/.test./, /.stories./];

const getOpeningElementName = path =>
  path.parent &&
  path.parent.openingElement &&
  path.parent.openingElement.name &&
  path.parent.openingElement.name.name;

const atrIsNotInBlackList = path => {
  const atrName =
    (path && path.parent && path.parent.name && path.parent.name.name) ||
    (path && path.parent && path.parent.key && path.parent.key.name);
  const parentAtrName =
    path &&
    path.parentPath &&
    path.parentPath.parent &&
    path.parentPath.parent.name &&
    path.parentPath.parent.name.name;

  const parentOfParentAtrName =
    path.parentPath &&
    path.parentPath.parentPath &&
    path.parentPath.parentPath.parent &&
    path.parentPath.parentPath.parent.name &&
    path.parentPath.parentPath.parent.name.name;

  const openingElementName = getOpeningElementName(path);

  return (
    !blackListOfAttributesToBeTransformed.includes(atrName) &&
    !blackListOfAttributesToBeTransformed.includes(parentAtrName) &&
    !blackListOfAttributesToBeTransformed.includes(parentOfParentAtrName) &&
    !blackListOfAttributesToBeTransformed.includes(openingElementName)
  );
};

const isPropertyKey = (path, value) => {
  if (path.parentPath.isProperty()) {
    return path.parent.key.value === value;
  }
  return false;
};

const isInWhiteListCallExpression = path => {
  const name =
    (path.parentPath.node.callee && path.parentPath.node.callee.name) ||
    (path.parentPath.node.callee &&
      path.parentPath.node.callee.property &&
      path.parentPath.node.callee.property.name);

  return whiteListOfCallExpressions.includes(name);
};

const isForbiddenParentType = path =>
  ((path.parentPath.isCallExpression() || path.parentPath.isNewExpression()) &&
    !isInWhiteListCallExpression(path)) ||
  path.parentPath.isTSEnumMember();

const isJSXAttribute = path =>
  path.parentPath.isJSXAttribute() ||
  path.parentPath.parentPath.isJSXAttribute();

const replaceWith = (t, path, replacement) => {
  const el = isJSXAttribute(path)
    ? t.jSXExpressionContainer(replacement)
    : replacement;

  if (path.parentPath.isJSXExpressionContainer()) {
    path.parentPath.replaceWith(el);
  } else {
    path.replaceWith(el);
  }
};

const isJSXElementWrapper = path =>
  path.parentPath &&
  path.parentPath.parentPath &&
  !path.parentPath.parentPath.isJSXAttribute() &&
  path.parentPath.parentPath.parentPath &&
  path.parentPath.parentPath.parentPath.isJSXElement();

const StringMod = t =>
  function(path, state) {
    const value = path.node.value && path.node.value.trim();
    const isOkey = isStringOkey(value);
    if (
      Boolean(value) &&
      isOkey &&
      atrIsNotInBlackList(path) &&
      !isPropertyKey(path, value) &&
      !isForbiddenParentType(path)
    ) {
      const replacement = isJSXElementWrapper(path)
        ? t.jsxElement(
            t.jsxOpeningElement(t.jsxIdentifier("Trans"), []),
            t.jsxClosingElement(t.jsxIdentifier("Trans")),
            [t.JSXText(value)]
          )
        : t.callExpression(
            t.memberExpression(t.Identifier("i18n"), t.Identifier("_"), false),
            [t.stringLiteral(value)]
          );

      replaceWith(t, path, replacement);
    }
  };

const getChildrenWithSpaces = (nodeChildren, t) => {
  const children = [];
  const nodeChildrenLength = nodeChildren.length;
  nodeChildren.forEach((el, index) => {
    if (t.isJSXText(el)) {
      const value = el.value;
      if (value[0] === " " && index !== 0) {
        children.push(t.JSXText(" "));
      }
      children.push(el);
      if (value[value.length - 1] === " " && index !== nodeChildrenLength - 1) {
        children.push(t.JSXText(" "));
      }
    } else {
      children.push(el);
    }
  });
  return children;
};

const getReplacementTemplateMod = (path, quasisValue, t) => {
  if (isJSXElementWrapper(path)) {
    const elements = [...path.node.quasis, ...path.node.expressions]
      .sort(
        ({ start: aStart = 0 } = {}, { start: bStart = 0 } = {}) =>
          aStart - bStart
      )
      .map(el => {
        if (t.isTemplateElement(el)) {
          return t.JSXText(el.value.cooked);
        }
        if (t.isIdentifier(el) || t.isMemberExpression(el)) {
          return t.jsxExpressionContainer(el);
        }
        return el;
      });

    return t.jsxElement(
      t.jsxOpeningElement(t.jsxIdentifier("Trans"), []),
      t.jsxClosingElement(t.jsxIdentifier("Trans")),
      getChildrenWithSpaces(elements, t)
    );
  } else {
    return t.callExpression(
      t.memberExpression(t.Identifier("i18n"), t.Identifier("_"), false),
      [
        path.node.expressions.length === 0
          ? t.stringLiteral(quasisValue)
          : t.taggedTemplateExpression(t.Identifier("t"), path.node)
      ]
    );
  }
};

const TemplateMod = t =>
  function(path, state) {
    const parent = path.parent;
    const quasisCookedArray = path.node.quasis.map(
      quasis => quasis.value.cooked
    );
    const quasisValue = quasisCookedArray.join("").trim();
    const isOkey = isStringOkey(quasisValue);

    if (
      !t.isTaggedTemplateExpression(parent) &&
      isOkey &&
      atrIsNotInBlackList(path) &&
      !isForbiddenParentType(path)
    ) {
      const replacement = getReplacementTemplateMod(path, quasisValue, t);

      path.replaceWith(replacement);
    }
  };

const getReplacementJSXMod = (path, t) => {
  const parentChildren = path.parent.children;
  const parentChildrenLength = parentChildren.length;
  if (getOpeningElementName(path) === "title") {
    const quasis = [];
    const expressions = [];
    const elements = parentChildren.map(el => {
      if (t.isJSXText(el)) {
        quasis.push(
          t.templateElement(
            { raw: el.value || "", cooked: el.value || "" },
            false
          )
        );
        return t.stringLiteral(el.value);
      }
      if (t.isJSXExpressionContainer(el)) {
        expressions.push(el.expression);
        return el.expression;
      }
      return el;
    });
    const children =
      parentChildrenLength > 1
        ? [
            t.taggedTemplateExpression(
              t.Identifier("t"),
              t.templateLiteral(quasis, expressions)
            )
          ]
        : elements;

    return t.jsxExpressionContainer(
      t.callExpression(
        t.memberExpression(t.Identifier("i18n"), t.Identifier("_"), false),
        children
      )
    );
  } else {
    const replacedPath =
      parentChildrenLength > 1
        ? getChildrenWithSpaces(path.parentPath.node.children, t)
        : [path.node];

    return t.jsxElement(
      t.jsxOpeningElement(t.jsxIdentifier("Trans"), []),
      t.jsxClosingElement(t.jsxIdentifier("Trans")),
      replacedPath
    );
  }
};

const JSXTextMod = (t, onlyComplex) =>
  function(path, state) {
    const value = path.node.value && path.node.value.trim();
    const childrenLength = path.parent.children.length;
    const parentHasTrans = path.findParent(p => {
      const openingElementName =
        p &&
        p.parent &&
        p.parent.openingElement &&
        p.parent.openingElement.name &&
        p.parent.openingElement.name.name;

      return openingElementName === "Trans";
    });

    const replaceOnleComplex = !onlyComplex || childrenLength > 1;
    const isOkey = isStringOkey(value, true);

    if (
      Boolean(value) &&
      isOkey &&
      atrIsNotInBlackList(path) &&
      !isPropertyKey(path, value) &&
      !isForbiddenParentType(path) &&
      !parentHasTrans &&
      replaceOnleComplex
    ) {
      const replacement = getReplacementJSXMod(path, t);

      if (childrenLength > 1) {
        const newParent = path.parentPath;
        newParent.node.children = [replacement];
        path.parentPath.replaceWith(newParent);
      } else {
        path.replaceWith(replacement);
      }
    }
  };

function stringLiteralReplace(_ref) {
  const t = _ref.types;

  return {
    manipulateOptions(opts, parserOpts) {
      parserOpts.plugins.push("classProperties");
      parserOpts.plugins.push("optionalChaining");
      parserOpts.plugins.push("objectRestSpread");
      parserOpts.plugins.push("nullishCoalescingOperator");
      parserOpts.plugins.push("numericSeparator");
    },
    visitor: {
      Program(path, opts) {
        if (blackListOfFileNames.some(regex => regex.test(opts.filename))) {
          return null;
        }
        let isTrans = false;
        let isI18n = false;
        let isMacro = false;
        let isLinguiCore = false;
        let isLinguiMacro = false;
        let isMacroTImported = false;
        let isMacroTransImported = false;

        const getMacroImport = (_isMacro, _isTrans) => {
          if (_isMacro) {
            isMacroTImported = true;
          }
          if (_isTrans) {
            isMacroTransImported = true;
          }
          return t.ImportDeclaration(
            [
              _isMacro &&
                t.ImportSpecifier(t.Identifier("t"), t.Identifier("t")),
              _isTrans &&
                t.ImportSpecifier(t.Identifier("Trans"), t.Identifier("Trans"))
            ].filter(Boolean),
            t.stringLiteral("@lingui/macro")
          );
        };

        path.traverse({
          TemplateLiteral: TemplateMod(t),
          StringLiteral: StringMod(t),
          JSXText: JSXTextMod(t, true)
        });

        path.traverse({
          JSXText: JSXTextMod(t, false)
        });

        path.traverse({
          JSXIdentifier(jsxPath) {
            if (jsxPath && jsxPath.node) {
              const name = jsxPath.node.name;
              if (name === "Trans") {
                isTrans = true;
              }
            }
          },
          MemberExpression(expPath) {
            if (expPath && expPath.node && expPath.node.object) {
              const name = expPath.node.object.name;
              if (name === "i18n") {
                isI18n = true;
              }
            }
          },
          TaggedTemplateExpression(taggedPath) {
            if (taggedPath && taggedPath.node && taggedPath.node.tag) {
              const name = taggedPath.node.tag.name;
              if (name === "t") {
                isMacro = true;
              }
            }
          },
          ImportDeclaration(importPath) {
            if (importPath && importPath.node && importPath.node.source) {
              const name = importPath.node.source.value;
              const specifiers = importPath.node.specifiers.map(
                specifier => specifier.local.name
              );
              if (name === "app/i18n") {
                isLinguiCore = true;
              }
              if (name === "@lingui/macro") {
                isLinguiMacro = true;
                if (specifiers.includes("t")) {
                  isMacroTImported = true;
                }
                if (specifiers.includes("Trans")) {
                  isMacroTransImported = true;
                }
              }
            }
          }
        });

        path.traverse({
          ImportDeclaration(importPath) {
            if (importPath && importPath.node && importPath.node.source) {
              const name = importPath.node.source.value;
              if (name === "@lingui/macro") {
                if (
                  (isMacro && !isMacroTImported) ||
                  (isTrans && !isMacroTransImported)
                ) {
                  importPath.replaceWith(getMacroImport(isMacro, isTrans));
                }
              }
            }
          }
        });

        if (isI18n && !isLinguiCore) {
          path.unshiftContainer(
            "body",
            t.ImportDeclaration(
              [t.ImportSpecifier(t.Identifier("i18n"), t.Identifier("i18n"))],
              t.stringLiteral("app/i18n")
            )
          );
        }
        if (!isLinguiMacro && (isMacro || isTrans)) {
          path.unshiftContainer(
            "body",
            t.ImportDeclaration(
              [
                isMacro &&
                  !isMacroTImported &&
                  t.ImportSpecifier(t.Identifier("t"), t.Identifier("t")),
                isTrans &&
                  !isMacroTransImported &&
                  t.ImportSpecifier(
                    t.Identifier("Trans"),
                    t.Identifier("Trans")
                  )
              ].filter(Boolean),
              t.stringLiteral("@lingui/macro")
            )
          );
        }
        return null;
      }
    }
  };
}

module.exports = stringLiteralReplace;
